//-----------------------------------------------------------------------------
//  Copyright (C) 2005-2015 Thomas S. Ullrich
//
//  This file is part of the Solitaire Mahjong project.
//
//  This file may be used under the terms of the GNU General Public License.
//  This project is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License.
//  
//  Author: Thomas S. Ullrich
//  Last update: April 8, 2014
//-----------------------------------------------------------------------------
#include <algorithm>
#include <iostream>
#include <QMessageBox>
#include <QHeaderView>
#include <QtDebug>
#include <QScreen>
#include <QGuiApplication>
#include "smPlayedTilesList.h"
#include "smDefinitions.h"
#include "smSettings.h"
#include "smTileItem.h"

smPlayedTilesList::smPlayedTilesList(QWidget* parent, Qt::WindowFlags fl) : QDialog(parent, fl)
{
    ui.setupUi(this);
    
    ui.mTableWidget->setSortingEnabled(false); // no sorting do it manually
    ui.mTableWidget->setAutoFillBackground(true);

    ui.mTableWidget->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    ui.mTableWidget->verticalHeader()->hide();
    ui.mTableWidget->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    ui.mTableWidget->horizontalHeader()->hide();
    
    ui.mTableWidget->setEditTriggers(QAbstractItemView::NoEditTriggers);
    
    connect(ui.mAsPlayedRadioButton, SIGNAL(toggled(bool)), this, SLOT(sortingModeHasChanged()));
    
    mSortAsPlayed = ui.mAsPlayedRadioButton->isChecked(); 
    
    QPalette thePalette(ui.mTableWidget->palette());
    thePalette.setColor(QPalette::Base, smSettings::instance()->mBackgroundColor);
    thePalette.setColor(QPalette::Window, smSettings::instance()->mBackgroundColor);
    ui.mTableWidget->setPalette(thePalette);
    
    move(smSettings::instance()->mPlayedTilesWindowPosition);
    resize(smSettings::instance()->mPlayedTilesWindowSize);
}

smPlayedTilesList::~smPlayedTilesList()
{ 
    smSettings::instance()->mPlayedTilesWindowPosition = pos();   
    smSettings::instance()->mPlayedTilesWindowSize = size();
}

void smPlayedTilesList::sortingModeHasChanged() 
{
    mSortAsPlayed = ui.mAsPlayedRadioButton->isChecked(); 
    layoutList();
}

void smPlayedTilesList::clear()
{
    ui.mTableWidget->clearContents();
}

void smPlayedTilesList::layoutList()
{
    ui.mTableWidget->clearContents();

    //
    //  Set same background color and tile size as on game board
    //
    QPalette thePalette(ui.mTableWidget->palette());
    thePalette.setColor(QPalette::Base, smSettings::instance()->mBackgroundColor);
    thePalette.setColor(QPalette::Window, smSettings::instance()->mBackgroundColor);
    ui.mTableWidget->setPalette(thePalette);
    
    //
    //  Layout as played, i.e. last one first always in pairs.
    //  This is essentially a dump of the history list.
    //
    if (mSortAsPlayed) {
        ui.mTableWidget->setColumnCount(2);
        ui.mTableWidget->setRowCount(static_cast<int>(mHistoryList.size()));
        for (unsigned int i=0; i<mHistoryList.size(); i++) {
            ui.mTableWidget->setItem(i, 0, new QTableWidgetItem(mHistoryList[i].first->mNormalPixmap,""));
            ui.mTableWidget->setItem(i, 1, new QTableWidgetItem(mHistoryList[i].second->mNormalPixmap,""));
        }
    }
    
    //
    //  Layout such that all that belong together are shown
    //  in one row. Needs some sorting. Use vector<vector> here.
    //
    else {  
        ui.mTableWidget->setColumnCount(4);
        vector<vector<smTileItem*> > sortedList = getSortedList(mHistoryList); 
        int nrows = 0; 
        for (unsigned int i=0; i<sortedList.size(); i++) 
            if (!sortedList[i].empty()) nrows++;
        ui.mTableWidget->setRowCount(nrows);
        nrows = 0;
        for (unsigned int i=0; i<sortedList.size(); i++) {
            for (unsigned int k=0; k<sortedList[i].size(); k++) {
                ui.mTableWidget->setItem(nrows, k, new QTableWidgetItem(sortedList[i][k]->mNormalPixmap,""));
            }
            if (!sortedList[i].empty()) nrows++;
        }
    }
    
    //
    // Set icon size (need for retina displays)
    //
    QScreen *screen = QGuiApplication::primaryScreen();

    if (mHistoryList.size()) {
        QSize size = mHistoryList[0].first->mNormalPixmap.size();
        size.setWidth(size.width()/screen->devicePixelRatio());
        size.setHeight(size.height()/screen->devicePixelRatio());
        ui.mTableWidget->setIconSize(size);
    }
    
    ui.mTableWidget->resizeColumnsToContents();
    ui.mTableWidget->resizeRowsToContents();
    update();
}

void smPlayedTilesList::layoutListQuick(pair<smTileItem*, smTileItem*> &last)
{
    //
    // No table traits are altered here. All we do is add one
    // pair. This method is not called when we switch from one
    // layout style to the other. This is handled by big brother
    // layoutList().
    //
    if (mSortAsPlayed) {
        ui.mTableWidget->insertRow(0);
        ui.mTableWidget->setItem(0, 0, new QTableWidgetItem(last.first->mNormalPixmap,""));
        ui.mTableWidget->setItem(0, 1, new QTableWidgetItem(last.second->mNormalPixmap,""));        
        ui.mTableWidget->setCurrentCell(0, 0);   // make sure the updated item is visible
    }
    else {
        vector<vector<smTileItem*> > sortedList = getSortedList(mHistoryList);
        int irow = 0;
        int actualRow = 0;
        bool erster = false;
        bool zweiter = false;
        for (unsigned int i=0; i<sortedList.size(); i++) {
            for (unsigned int k=0; k<sortedList[i].size(); k++) {
                if (last.first == sortedList[i][k])  {
                    irow = i; 
                    erster = true;
                }
                if (last.second == sortedList[i][k]) zweiter = true;
                if (erster && zweiter) break;
            }
            if (erster && zweiter) break;
            if (!sortedList[i].empty()) actualRow++;
        }
        if (!(erster && zweiter)) {  // something wrong, let big brother do it
            layoutList();
            return;
        }
 
        if (sortedList[irow].size() == 2) {  // new row needed
            ui.mTableWidget->insertRow(actualRow);
            ui.mTableWidget->setItem(actualRow, 0, new QTableWidgetItem(last.first->mNormalPixmap,""));
            ui.mTableWidget->setItem(actualRow, 1, new QTableWidgetItem(last.second->mNormalPixmap,""));
        }
        else {
            ui.mTableWidget->setItem(actualRow, 2, new QTableWidgetItem(last.first->mNormalPixmap,""));
            ui.mTableWidget->setItem(actualRow, 3, new QTableWidgetItem(last.second->mNormalPixmap,""));
        }
        ui.mTableWidget->setCurrentCell(actualRow, 0);  // make sure the updated item is visible
    }
    ui.mTableWidget->update();
}

vector<vector<smTileItem*> > smPlayedTilesList::getSortedList(const vector<pair<smTileItem*, smTileItem*> > &list) 
{
    vector<vector<smTileItem*> > sortedList(Dragon_White+2+1); // see smDefinitions.h
    for (unsigned int i=0; i<list.size(); i++) {
        int index = 0;
        if (list[i].first->mTileId <= Dragon_White) 
            index = list[i].first->mTileId;
        else if (list[i].first->mTileId >= Season_1 && list[i].first->mTileId <= Season_4)
            index = Dragon_White+1;
        else if (list[i].first->mTileId >= Flower_1 && list[i].first->mTileId <= Flower_4)
            index = Dragon_White+2;
        sortedList[index].push_back(list[i].first);                
        sortedList[index].push_back(list[i].second);
    }  
    return sortedList;
}

void smPlayedTilesList::sync(stack<pair<smTileItem*, smTileItem*> > historyList)
{
    //
    // Since rebuilding the table takes a long time if displayed
    // permanently we check here if only one pair is to added. 
    // Except for undos and some perculiar cases this is the norm.
    // In this case we call a specialized layout version. 
    // If things are more complicated we start from scratch and 
    // use layoutList()
    //
    vector<pair<smTileItem*, smTileItem*> > tmpHistoryVector;
    while (!historyList.empty()) {
        tmpHistoryVector.push_back(historyList.top());
        historyList.pop();
    }
    
    bool identicalExceptLast = true;
    if (tmpHistoryVector.size() == mHistoryList.size()+1) {
        for (unsigned int i=0; i<mHistoryList.size(); i++) {
            if (mHistoryList[i] != tmpHistoryVector[i+1]) {
                identicalExceptLast = false;
                break;
            }
        }
    }
    else {
        identicalExceptLast = false;
    }
    
    mHistoryList = tmpHistoryVector;

    if (identicalExceptLast && mHistoryList.size() > 2) 
        layoutListQuick(mHistoryList[0]);
    else
        layoutList();
 }

void smPlayedTilesList::hideEvent(QHideEvent*)
{
    smSettings::instance()->mPlayedTilesWindowPosition = pos();   
    smSettings::instance()->mPlayedTilesWindowSize = size();   
}


void smPlayedTilesList::showEvent(QShowEvent*)
{
    move(smSettings::instance()->mPlayedTilesWindowPosition);
    resize(smSettings::instance()->mPlayedTilesWindowSize);
}